#ifndef VY_VOYAGE_H
#define VY_VOYAGE_H

/* basic types and macros */

#include <stddef.h>
#include <stdint.h>

#if defined(_MSC_VER) && !defined(VY_STATIC_LIB)
#define VY_DLLEXPORT __declspec(dllexport)
#else
#define VY_DLLEXPORT
#endif

#if defined(_MSC_VER)
#define VY_INLINE __forceinline
#elif defined(__GNUC__) || defined(__clang__)
#define VY_INLINE inline __attribute__((always_inline))
#endif


#define VY_UNUSED(x)      ((void)sizeof((x)))
#define VY_ARRAY_COUNT(x) (sizeof((x)) / sizeof((x)[0]))

#define VY_KB(n) ((n)*1024U)
#define VY_MB(n) ((n)*1024U * 1024U)
#define VY_GB(n) ((n)*1024U * 1024U * 1024U)

typedef unsigned int vy_result;
#define VY_SUCCESS 0
#define VY_UNKNOWN_ERROR (vy_result)-1

typedef struct {
    const char *start;
    unsigned int length;
} vy_text_span;

/* Returns results like strcmp():
 * - If the first differing character is smaller in span than in cmp: < 0
 * - If span and cmp are equal: 0
 * - If the first differing character is greater in span than in cmp: > 0
 * - If span.length is smaller than strlen(cmp): -1
 * - If span.length is greater than strlen(cmp): 1
 */
VY_DLLEXPORT int vyCompareSpanToString(vy_text_span span, const char *cmp);

VY_DLLEXPORT void vyReportError(const char *subsystem, const char *fmt, ...);

VY_DLLEXPORT void vyLog(const char *subsystem, const char *fmt, ...);

enum {
    VY_INVALID_UNICODE     = 1,
    VY_INSUFFICIENT_BUFFER = 2,
};

/* Returns VY_SUCCESS if the string was successfully converted,
 * VY_INVALID_UNICODE if invalid unicode characters were encountered or 
 * VY_INSUFFICIENT_BUFFER if the provided output buffer is too small */
VY_DLLEXPORT vy_result vyUTF8ToWStr(const char *utf8, wchar_t *wstr, size_t len);

/* Returns VY_SUCCESS if the string was successfully converted,
 * VY_INVALID_UNICODE if invalid unicode characters were encountered or
 * VY_INSUFFICIENT_BUFFER if the provided output buffer is too small */
VY_DLLEXPORT vy_result vyWStrToUTF8(const wchar_t *wstr, char *utf8, size_t len);

/* Relative pointer */
typedef struct {
    int32_t off;
} vy_relptr;

static VY_INLINE void *vyResolveRelptr(vy_relptr *ptr) {
    if (ptr->off != 0) {
        char *p = (char *)ptr;
        return (void *)(p + (ptrdiff_t)ptr->off);
    } else {
        return NULL;
    }
}

static VY_INLINE void vySetRelptr(vy_relptr *ptr, void *target) {
    if (target) {
        char *p = (char *)ptr, *t = (char *)target;
        ptrdiff_t d = t - p;
        ptr->off    = (int32_t)d;
    } else {
        ptr->off = 0;
    }
}

/* Bitfiddling functions */

/* Portable solution, On x64 using clzl is probably faster. */
static VY_INLINE uint32_t vyNextPowerOfTwo32(uint32_t v) {
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    return v;
}


/* Runtime init. Initializes basic systems.
 * You need to call this, even if you build a CLI only app. */
VY_DLLEXPORT vy_result vyInitRuntime(void);

VY_DLLEXPORT void vyShutdownRuntime(void);

#endif
