#include "assetsettings.h"
#include "description_parser.h"
#include "packages.h"
#include "utils.h"

#include "runtime/aio.h"
#include "runtime/buffer_manager.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

vy_result vyParseAssetSettings(const char *text,
                               size_t length,
                               const char *file_path,
                               vy_asset_settings *settings) {
    unsigned int root_list;
    vy_parse_state state;
    vy_result res = vyParseDescription(text, length, file_path, &root_list, &state);
    if (res != VY_SUCCESS) {
        vyReportError("ASSETC", "Failed to parse asset settings: %s", file_path);
        return res;
    }

    /* Default settings */
    settings->package = 0;
    settings->processing_flags = 0;
    settings->reprocess_on_dependency_change = false;
    
    const vy_parsed_stmt *package_stmt = vyFindStatement(&state, root_list, "package");
    if (package_stmt) {
        if (package_stmt->form != VY_STMT_FORM_VALUE) {
            vyReportError("ASSETC",
                          "Expected a package name as the value of 'package' in %s.",
                          file_path);
            res = VY_UNKNOWN_ERROR;
            goto out;
        }
        settings->package = vyAddPackageFile(package_stmt->value);
    }

    const vy_parsed_stmt *flags_stmt = vyFindStatement(&state, root_list, "processing_flags");
    if (flags_stmt) {
        if (flags_stmt->form != VY_STMT_FORM_VALUE) {
            vyReportError("ASSETC",
                          "Expected a hexadecimal number as the value of 'processing_flags' in %s.",
                          file_path);
            res = VY_UNKNOWN_ERROR;
            goto out;
        }

        sscanf(flags_stmt->value.start, "%x", &settings->processing_flags);
    }

    const vy_parsed_stmt *reprocess_stmt =
        vyFindStatement(&state, root_list, "reprocess_on_dependency_change");
    if (reprocess_stmt) {
        if (reprocess_stmt->form != VY_STMT_FORM_VALUE) {
            vyReportError("ASSETC",
                          "Expected either 'true' or 'false' as the value of 'reprocess_on_dependency_change' in %s.",
                          file_path);
            res = VY_UNKNOWN_ERROR;
            goto out;
        }
        if (vyCompareSpanToString(reprocess_stmt->value, "true") == 0)
            settings->reprocess_on_dependency_change = true;
        else if (vyCompareSpanToString(reprocess_stmt->value, "false") == 0)
            settings->reprocess_on_dependency_change = false;
        else {
            vyReportError("ASSETC",
                          "Expected either 'true' or 'false' as the value of "
                          "'reprocess_on_dependency_change' in %s.",
                          file_path);
            res = VY_UNKNOWN_ERROR;
            goto out;
        }
    }

out:
    vyReleaseParseState(&state);
    return res;
}

vy_result vyLoadAssetSettings(const char *asset_path, vy_asset_settings *settings) {
    size_t path_len = strlen(asset_path);

    char *as_path = malloc(path_len + 3);
    if (!as_path) {
        return VY_UNKNOWN_ERROR;
    }
    memcpy(as_path, asset_path, path_len);

    size_t ext_len = 0;
    while (ext_len < path_len) {
        if (as_path[path_len - ext_len] == '.')
            break;
        ++ext_len;
    }

    strcpy(&as_path[path_len - ext_len], ".as");
    size_t as_size = vyGetFileSize(as_path);
    if (as_size == 0) {
        vyReportError("ASSETC", "Failed to retrieve size of setting file %s", as_path);
        free(as_path);
        return VY_UNKNOWN_ERROR;
    }

    void *as_buffer = vyAllocBuffer(as_size);

    vy_load_batch as_load;
    as_load.loads[0].file      = vyAddFile(as_path);
    as_load.loads[0].num_bytes = as_size;
    as_load.loads[0].dest      = as_buffer;
    if (!as_load.loads[0].dest) {
        vyReportError("ASSETC", "Failed to allocate buffer for setting file %s", as_path);
        free(as_path);
        return VY_UNKNOWN_ERROR;
    }
    as_load.loads[0].offset = 0;
    as_load.num_loads       = 1;

    vy_aio_handle as_handle;
    if (vySubmitLoadBatch(&as_load, &as_handle) != VY_SUCCESS) {
        vyReportError("ASSETC", "Failed to submit load of setting file %s", as_path);
        free(as_path);
        return VY_UNKNOWN_ERROR;
    }

    if (vyWaitForAIOCompletion(as_handle) != VY_AIO_STATE_FINISHED) {
        vyReportError("ASSETC", "Failed to load setting file %s", as_path);
        free(as_path);
        return VY_UNKNOWN_ERROR;
    }

    vyReleaseAIO(as_handle);

    if (vyParseAssetSettings(as_buffer, as_size, as_path, settings) != VY_SUCCESS) {
        free(as_path);
        return VY_UNKNOWN_ERROR;
    }

    free(as_path);
    return VY_SUCCESS;
}