#include <assert.h>
#include <limits.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "runtime/aio.h"
#include "runtime/gfx.h"
#include "runtime/handles.h"
#include "runtime/runtime.h"
#include "runtime/buffer_manager.h"
#include "runtime/renderer_api.h"

#include "assetmeta.h"
#include "description_parser.h"
#include "options.h"
#include "processing.h"
#include "utils.h"
#include "processing_flags.h"
#include "dependency_tracking.h"

typedef struct {
    vy_attribute_binding *uniform_bindings;
    vy_attribute_binding *storage_bindings;
    vy_attribute_binding *texture_bindings;

    vy_uid vertex_shader;
    vy_uid fragment_shader;
    vy_uid compute_shader;

    /* TODO Fixed function settings */

    /* Sampler settings */

    uint16_t uniform_binding_count;
    uint16_t storage_binding_count;
    uint16_t texture_binding_count;
} vy_parsed_pipeline_data;

static void
DbgPrintShaderFile(const vy_parse_state *state, unsigned int list_index, unsigned int indent) {
    assert(list_index < state->statement_list_count);
    const vy_parsed_stmt_list *list = &state->statement_lists[list_index];

    unsigned int stmt_index = list->first;
    for (unsigned int i = 0; i < list->count; ++i) {
        const vy_parsed_stmt *stmt = &state->statements[stmt_index];
        for (unsigned int j = 0; j < indent; ++j)
            printf(" ");
        printf("%.*s: ", stmt->attribute.length, stmt->attribute.start);
        if (stmt->form == VY_STMT_FORM_VALUE) {
            printf("%.*s\n", stmt->value.length, stmt->value.start);
        } else {
            printf("{\n");
            DbgPrintShaderFile(state, stmt->list_index, indent + 2);
            printf("}\n");
        }
        stmt_index = stmt->next;
    }
    assert(stmt_index == UINT_MAX || stmt_index == 0);
}

static bool ParseBindingIndex(vy_text_span span, unsigned int *index) {
    if (span.length == 0)
        return false;
    int at           = (int)span.length - 1;
    unsigned int exp = 1;
    unsigned int n   = 0;
    while (at >= 0) {
        if (span.start[at] >= '0' && span.start[at] <= '9') {
            unsigned int digit = (unsigned int)(span.start[at] - '0');
            n += digit * exp;
        } else {
            vyReportError("GFX", "Unexpected non-digit character in binding index");
            return false;
        }
        --at;
        exp *= 10;
    }
    *index = n;
    return true;
}

static vy_attribute_value ParseBindingValue(vy_text_span span) {
    if (vyCompareSpanToString(span, "MATERIAL_ALBEDO") == 0) {
        return VY_ATTRIBUTE_VALUE_MATERIAL_ALBEDO;
    } else if (vyCompareSpanToString(span, "MATERIAL_NORMAL") == 0) {
        return VY_ATTRIBUTE_VALUE_MATERIAL_NORMAL;
    }
    vyReportError("GFX", "Unsupported binding value %*.s", span.length, span.start);
    return VY_ATTRIBUTE_VALUE_UNDEFINED;
}

static bool ParseBindings(vy_parse_state *state,
                          unsigned int root_list,
                          const char *name,
                          const char *file_path,
                          vy_attribute_binding **p_bindings,
                          uint16_t *p_binding_count) {
    const vy_parsed_stmt *bindings = vyFindStatement(state, root_list, name);
    if (bindings) {
        if (bindings->form != VY_STMT_FORM_LIST) {
            vyReportError("GFX",
                          "Expected list of bindings as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return false;
        }
        const vy_parsed_stmt_list *binding_list = &state->statement_lists[bindings->list_index];
        vy_attribute_binding *shader_bindings =
            vyAllocBuffer(sizeof(vy_attribute_binding) * binding_list->count);
        if (!bindings) {
            vyReportError("GFX", "Out of memory");
            return false;
        }
        unsigned int binding_count = binding_list->count;

        unsigned int stmt_index = binding_list->first;
        for (unsigned int i = 0; i < binding_list->count; ++i) {
            const vy_parsed_stmt *stmt = &state->statements[stmt_index];
            if (!ParseBindingIndex(stmt->attribute, &shader_bindings[i].index)) {
                vyReleaseBuffer(shader_bindings,
                                sizeof(vy_attribute_binding) * binding_list->count);
                return false;
            }

            shader_bindings[i].value = ParseBindingValue(stmt->value);
            if (shader_bindings[i].value == VY_ATTRIBUTE_VALUE_UNDEFINED) {
                vyReleaseBuffer(shader_bindings,
                                sizeof(vy_attribute_binding) * binding_list->count);
                return false;
            }
            stmt_index = stmt->next;
        }

        *p_bindings      = shader_bindings;
        *p_binding_count = (uint16_t)binding_count;
        return true;
    } else {
        *p_bindings      = NULL;
        *p_binding_count = 0;
        return true;
    }
}

static vy_result ParseShader(vy_parse_state *state,
                             unsigned int root_list,
                             const char *name,
                             const char *file_path,
                             uint32_t processing_flags,
                             vy_uid *p_shader_uid) {
    const vy_parsed_stmt *stmt = vyFindStatement(state, root_list, name);
    if (stmt) {
        if (stmt->form != VY_STMT_FORM_LIST) {
            vyReportError("GFX",
                          "Expected a list as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return VY_PROCESSING_FAILED;
        }
        const vy_parsed_stmt_list *shader_list = &state->statement_lists[stmt->list_index];

        unsigned int stmt_index = shader_list->first;
        for (unsigned int i = 0; i < shader_list->count; ++i) {
            const vy_parsed_stmt *shader = &state->statements[stmt_index];
            if (shader->form != VY_STMT_FORM_VALUE) {
                vyReportError("GFX",
                              "Expected a list as the value of "
                              "\"%s.%*s\" in %s",
                              name,
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              file_path);
                return VY_PROCESSING_FAILED;
            }
            vy_renderer_backend_code backend = VY_INVALID_RENDERER_BACKEND_CODE;
            if (vyCompareSpanToString(shader->attribute, "vk") == 0) {
                backend = VY_RENDERER_BACKEND_CODE_VK;
            } else {
                vyReportError("GFX",
                              "Invalid renderer backend"
                              "\"%*s\" in %s of file %s",
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              name,
                              file_path);
                return VY_PROCESSING_FAILED;
            }

            if (backend == g_assetc_options.renderer_backend) {
                vy_file_id shader_file = vyAddFileFromSpan(shader->value);
                vy_uid uid             = vyLookupUID(shader_file);
                if (uid == VY_INVALID_UID) {
                    /* Add the shader file to processing and wait until its done
                     */
                    if (vyAddFileToProcessingQueue(shader_file, processing_flags) != VY_SUCCESS)
                        return VY_PROCESSING_FAILED;
                    return VY_PROCESSING_TRY_AGAIN;
                }
                *p_shader_uid = uid;

                /* Don't break, because that validates the file. (attributes are checked for invalid
                 * values). */
            }
            stmt_index = shader->next;
        }
        return VY_SUCCESS;
    }
    return VY_PROCESSING_FAILED;
}

static uint32_t
ParseOptimizationLevel(vy_parse_state *state, unsigned int root_list, const char *file_path) {
    uint32_t optimization_level;
    switch (g_assetc_options.optimization) { 
    case VY_ASSET_OPTIMIZATION_PERFORMANCE:
        optimization_level = VY_SHADER_FLAG_OPTIMIZE_SPEED;
        break;
    case VY_ASSET_OPTIMIZATION_SPACE:
        optimization_level = VY_SHADER_FLAG_OPTIMIZE_SIZE;
        break;
    default:
        optimization_level = 0;
    }

    const vy_parsed_stmt *stmt = vyFindStatement(state, root_list, "optimization");

    if (stmt) {
        if (stmt->form != VY_STMT_FORM_VALUE) {
            vyReportError("GFX",
                          "Expected a simple statement for"
                          "\"optimization\" in %s",
                          file_path);
            return optimization_level;
        }

        if (vyCompareSpanToString(stmt->value, "speed") == 0) {
            optimization_level = VY_SHADER_FLAG_OPTIMIZE_SPEED;
        } else if (vyCompareSpanToString(stmt->value, "size") == 0) {
            optimization_level = VY_SHADER_FLAG_OPTIMIZE_SIZE;
        } else if (vyCompareSpanToString(stmt->value, "none") == 0) {
            optimization_level = 0;
        } else {
            vyReportError("GFX",
                          "Expected one of 'speed', 'size' and 'none' for \"optimization\" in %s",
                          file_path);
        }
    }

    return optimization_level;
}

static vy_result
ParsePipelineFile(vy_file_id fid, const char *text, size_t length, vy_parsed_pipeline_data *pipeline) {
    /* This is the grammar for pipeline files:
     * <stmt-list> ::= <stmt>*
     * <stmt> ::= <attribute> ( ( <value> ';' ) | ( '{' <stmt-list> '}' ) )
     * <attribute> ::= [:alnum:]*
     * <value>:: = [:alnum:]* */
    const char *file_path = vyGetFilePath(fid);
    vy_parse_state state;
    unsigned int root_list;
    vy_result result = vyParseDescription(text, length, file_path, &root_list, &state);
    if (result != VY_SUCCESS) {
        goto out;
    }

    /* We allow the pipeline file to overwrite the optimization level */
    uint32_t optimization = ParseOptimizationLevel(&state, root_list, file_path);

    /* Process shader stages */
    if (ParseShader(&state,
                    root_list,
                    "vertex",
                    file_path,
                    VY_SHADER_FLAG_VERTEX | optimization,
                    &pipeline->vertex_shader) == VY_PROCESSING_TRY_AGAIN) {
        result = VY_PROCESSING_TRY_AGAIN;
        goto out;
    }
    if (ParseShader(&state,
                    root_list,
                    "fragment",
                    file_path,
                    VY_SHADER_FLAG_FRAGMENT | optimization,
                    &pipeline->fragment_shader) == VY_PROCESSING_TRY_AGAIN) {
        result = VY_PROCESSING_TRY_AGAIN;
        goto out;
    }
    if (ParseShader(&state,
                    root_list,
                    "compute",
                    file_path,
                    VY_SHADER_FLAG_COMPUTE | optimization,
                    &pipeline->compute_shader) == VY_PROCESSING_TRY_AGAIN) {
        result = VY_PROCESSING_TRY_AGAIN;
        goto out;
    }

    /* Process bindings */
    pipeline->texture_bindings      = NULL;
    pipeline->texture_binding_count = 0;
    pipeline->uniform_bindings      = NULL;
    pipeline->uniform_binding_count = 0;
    pipeline->storage_bindings      = NULL;
    pipeline->storage_binding_count = 0;

    if (!ParseBindings(&state,
                       root_list,
                       "texture_bindings",
                       file_path,
                       &pipeline->texture_bindings,
                       &pipeline->texture_binding_count)) {
        result = VY_AIO_STATE_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "uniform_bindings",
                       file_path,
                       &pipeline->uniform_bindings,
                       &pipeline->uniform_binding_count)) {
        result = VY_AIO_STATE_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "storage_bindings",
                       file_path,
                       &pipeline->storage_bindings,
                       &pipeline->storage_binding_count)) {
        result = VY_AIO_STATE_FAILED;
        goto out;
    }

out:
    free(state.statements);
    free(state.statement_lists);
    return result;
}

vy_result vyProcessPipelineFile(vy_file_id file,
                                void *buffer,
                                size_t size,
                                uint32_t flags,
                                vy_processor_output *output) {
    VY_UNUSED(flags);
    vy_parsed_pipeline_data tmp;
    memset(&tmp, 0, sizeof(tmp));

    vy_result result = ParsePipelineFile(file, buffer, size, &tmp);
    if (result == VY_SUCCESS) {
        /* parsed_pipeline_data contains arrays of bindings.
         * We need to convert these to a flat buffer that can be written to a file */

        size_t outbuffer_size =
            sizeof(vy_pipeline_info) +
            sizeof(vy_attribute_binding) *
                (tmp.storage_binding_count + tmp.texture_binding_count + tmp.uniform_binding_count);
        void *out_buffer = vyAllocBuffer(outbuffer_size);
        if (!buffer) {
            return VY_PROCESSING_FAILED;
        }
        vy_pipeline_info *info      = out_buffer;
        info->vertex_shader         = tmp.vertex_shader;
        info->fragment_shader       = tmp.fragment_shader;
        info->compute_shader        = tmp.compute_shader;
        info->texture_binding_count = tmp.texture_binding_count;
        info->uniform_binding_count = tmp.uniform_binding_count;
        info->storage_binding_count = tmp.storage_binding_count;

        vy_attribute_binding *texture_bindings = NULL;
        if (tmp.texture_binding_count > 0) {
            texture_bindings = (vy_attribute_binding *)(info + 1);
            memcpy(texture_bindings,
                   tmp.texture_bindings,
                   sizeof(vy_attribute_binding) * tmp.texture_binding_count);
        }
        vy_attribute_binding *uniform_bindings = NULL;
        if (tmp.uniform_binding_count > 0) {
            uniform_bindings = texture_bindings + tmp.texture_binding_count;
            memcpy(uniform_bindings,
                   tmp.uniform_bindings,
                   sizeof(vy_attribute_binding) * tmp.uniform_binding_count);
        }
        vy_attribute_binding *storage_bindings = NULL;
        if (tmp.storage_binding_count > 0) {
            storage_bindings = uniform_bindings + tmp.uniform_binding_count;
            memcpy(storage_bindings,
                   tmp.storage_bindings,
                   sizeof(vy_attribute_binding) * tmp.storage_binding_count);
        }
        vySetRelptr(&info->texture_bindings, texture_bindings);
        vySetRelptr(&info->uniform_bindings, uniform_bindings);
        vySetRelptr(&info->storage_bindings, storage_bindings);

        vyReleaseBuffer(tmp.texture_bindings,
                        sizeof(vy_attribute_binding) * tmp.texture_binding_count);
        vyReleaseBuffer(tmp.storage_bindings,
                        sizeof(vy_attribute_binding) * tmp.storage_binding_count);
        vyReleaseBuffer(tmp.uniform_bindings,
                        sizeof(vy_attribute_binding) * tmp.uniform_binding_count);

        output->data      = out_buffer;
        output->size      = outbuffer_size;
        output->asset_uid = vyCalculateUID(vyGetFilePath(file));

        /* Store dependencies to shaders */
        if (info->vertex_shader != VY_INVALID_UID)
            result = vyAddAssetDependency(output->asset_uid, info->vertex_shader);
        if (info->fragment_shader != VY_INVALID_UID)
            result = vyAddAssetDependency(output->asset_uid, info->fragment_shader);
        if (info->compute_shader != VY_INVALID_UID)
            result = vyAddAssetDependency(output->asset_uid, info->compute_shader);
    }
    return result;
}